import configparser
import json
from abc import ABC
from urllib.parse import quote_plus


class BaseConfig(ABC):

    def __init__(self, config_file='dicebox.config'):
        ###############################################################################
        # Create config objects.
        ###############################################################################
        local_config = configparser.ConfigParser()
        local_config.read(config_file)

        ###############################################################################
        # Data Set Options
        ###############################################################################

        # Load user defined config
        self.DATASET = local_config.get('DATASET', 'name')
        self.DICEBOX_COMPLIANT_DATASET = local_config.getboolean('DATASET', 'dicebox_compliant')
        self.NB_CLASSES = local_config.getint('DATASET', 'categories')
        self.IMAGE_WIDTH = local_config.getint('DATASET', 'image_width')
        self.IMAGE_HEIGHT = local_config.getint('DATASET', 'image_height')
        self.DATA_BASE_DIRECTORY = local_config.get('DIRECTORY', 'dataset_base_directory')

        # Build Calculated Configs
        self.NETWORK_NAME = "%s_%ix%i" % (self.DATASET, self.IMAGE_WIDTH, self.IMAGE_HEIGHT)
        self.INPUT_SHAPE = (self.IMAGE_WIDTH, self.IMAGE_HEIGHT, 3)  # 3 indicates the number of channels (RGB) so 3.
        self.DATA_DIRECTORY = "%s/%s/data/" % (self.DATA_BASE_DIRECTORY, self.NETWORK_NAME)

        ###############################################################################
        # Neural Network Taxonomy Options
        ###############################################################################
        self.MIN_NEURONS = local_config.getint('TAXONOMY', 'min_neurons')
        self.MAX_NEURONS = local_config.getint('TAXONOMY', 'max_neurons')
        self.MIN_LAYERS = local_config.getint('TAXONOMY', 'min_layers')
        self.MAX_LAYERS = local_config.getint('TAXONOMY', 'max_layers')
        self.LAYER_TYPES = local_config.get('TAXONOMY', 'layer_types')
        self.ACTIVATION = local_config.get('TAXONOMY', 'activation')
        self.OPTIMIZER = local_config.get('TAXONOMY', 'optimizer')

        self.TAXONOMY = {
            'min_neurons': self.MIN_NEURONS,
            'max_neurons': self.MAX_NEURONS,
            'min_layers': self.MIN_LAYERS,
            'max_layers': self.MAX_LAYERS,
            'layer_types': json.loads(self.LAYER_TYPES),
            'activation': json.loads(self.ACTIVATION),
            'optimizer': json.loads(self.OPTIMIZER)
        }

        ###############################################################################
        # Evolution Options
        ###############################################################################
        self.EPOCHS = local_config.getint('EVOLUTION', 'epochs')

        # Number of times to evole the population.
        self.GENERATIONS = local_config.getint('EVOLUTION', 'generations')

        # Number of networks in each generation.
        self.POPULATION = local_config.getint('EVOLUTION', 'population')
        self.NOISE = local_config.getfloat('GLOBAL', 'noise')

        ###############################################################################
        # Training Options / Settings for the 1920x1080 dataset
        ###############################################################################
        self.BATCH_SIZE = local_config.getint('TRAINING', 'batch_size')
        self.TRAIN_BATCH_SIZE = local_config.getint('TRAINING', 'train_batch_size')
        self.TEST_BATCH_SIZE = local_config.getint('TRAINING', 'test_batch_size')
        self.LOAD_BEST_WEIGHTS_ON_START = local_config.getboolean('TRAINING', 'load_best_weights_on_start')

        ###############################################################################
        # Direcrtory Options
        ###############################################################################
        self.LOGS_DIR = local_config.get('DIRECTORY', 'logs_dir')
        self.WEIGHTS_DIR = local_config.get('DIRECTORY', 'weights_dir')
        self.TMP_DIR = local_config.get('DIRECTORY', 'tmp_dir')
        self.POPULATION_DIR = local_config.get('DIRECTORY', 'population_dir')

        ###############################################################################
        # Server Options
        ###############################################################################
        self.API_ACCESS_KEY = local_config.get('SERVER', 'api_access_key')
        self.API_VERSION = local_config.get('SERVER', 'api_version')
        self.LISTENING_HOST = local_config.get('SERVER', 'listening_host')
        self.FLASK_DEBUG = local_config.getboolean('SERVER', 'flask_debug')
        self.MODEL_WEIGHTS_FILENAME = local_config.get('SERVER', 'model_weights_filename')

        ###############################################################################
        # Sensory Service Options
        ###############################################################################
        self.SENSORY_SERVER = local_config.get('SENSORY_SERVICE', 'sensory_server')
        self.SENSORY_PORT = local_config.getint('SENSORY_SERVICE', 'sensory_port')
        self.SENSORY_URI = local_config.get('SENSORY_SERVICE', 'sensory_uri')

        self.SENSORY_SERVICE_RABBITMQ_EXCHANGE = local_config.get('SENSORY_SERVICE', 'rabbitmq_exchange')
        self.SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_ROUTING_KEY = local_config.get(
            'SENSORY_SERVICE', 'rabbitmq_batch_request_routing_key')
        self.SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_TASK_QUEUE = local_config.get(
            'SENSORY_SERVICE', 'rabbitmq_batch_request_task_queue')

        self.SENSORY_SERVICE_RABBITMQ_URI = local_config.get('SENSORY_SERVICE', 'rabbitmq_uri')
        self.SENSORY_SERVICE_RABBITMQ_USERNAME = local_config.get('SENSORY_SERVICE', 'rabbitmq_username')
        self.SENSORY_SERVICE_RABBITMQ_PASSWORD = local_config.get('SENSORY_SERVICE', 'rabbitmq_password')
        self.SENSORY_SERVICE_RABBITMQ_SERVER = local_config.get('SENSORY_SERVICE', 'rabbitmq_server')
        self.SENSORY_SERVICE_RABBITMQ_PORT = local_config.getint('SENSORY_SERVICE', 'rabbitmq_port')
        self.SENSORY_SERVICE_RABBITMQ_VHOST = quote_plus(
            local_config.get('SENSORY_SERVICE', 'rabbitmq_vhost'))

        self.SENSORY_SERVICE_RABBITMQ_URL = "%s%s:%s@%s:%s/%s" % (
            self.SENSORY_SERVICE_RABBITMQ_URI,
            self.SENSORY_SERVICE_RABBITMQ_USERNAME,
            self.SENSORY_SERVICE_RABBITMQ_PASSWORD,
            self.SENSORY_SERVICE_RABBITMQ_SERVER,
            self.SENSORY_SERVICE_RABBITMQ_PORT,
            self.SENSORY_SERVICE_RABBITMQ_VHOST
        )
        self.SENSORY_SERVICE_SHARD_SIZE = local_config.getint('SENSORY_SERVICE', 'shard_size')

        ###############################################################################
        # Training Service Options
        ###############################################################################
        self.TRAINING_SERVICE_RABBITMQ_EXCHANGE = local_config.get(
            'TRAINING_SERVICE', 'rabbitmq_exchange')
        self.TRAINING_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY = local_config.get(
            'TRAINING_SERVICE', 'rabbitmq_batch_request_routing_key')
        self.TRAINING_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE = local_config.get(
            'TRAINING_SERVICE', 'rabbitmq_train_request_task_queue')
        self.TRAINING_SERVICE_RABBITMQ_RABBITMQ_URI = local_config.get('TRAINING_SERVICE', 'rabbitmq_uri')
        self.TRAINING_SERVICE_RABBITMQ_USERNAME = local_config.get('TRAINING_SERVICE', 'rabbitmq_username')
        self.TRAINING_SERVICE_RABBITMQ_PASSWORD = local_config.get('TRAINING_SERVICE', 'rabbitmq_password')
        self.TRAINING_SERVICE_RABBITMQ_SERVER = local_config.get('TRAINING_SERVICE', 'rabbitmq_server')
        self.TRAINING_SERVICE_RABBITMQ_PORT = local_config.getint('TRAINING_SERVICE', 'rabbitmq_port')
        self.TRAINING_SERVICE_RABBITMQ_VHOST = quote_plus(local_config.get('TRAINING_SERVICE', 'rabbitmq_vhost'))
        self.TRAINING_SERVICE_RABBITMQ_URL = "%s%s:%s@%s:%s/%s" % (
            self.TRAINING_SERVICE_RABBITMQ_RABBITMQ_URI,
            self.TRAINING_SERVICE_RABBITMQ_USERNAME,
            self.TRAINING_SERVICE_RABBITMQ_PASSWORD,
            self.TRAINING_SERVICE_RABBITMQ_SERVER,
            self.TRAINING_SERVICE_RABBITMQ_PORT,
            self.TRAINING_SERVICE_RABBITMQ_VHOST
        )

        ###############################################################################
        # Training Processor Options
        ###############################################################################
        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_EXCHANGE = local_config.get(
            'TRAINING_PROCESSOR_SERVICE', 'rabbitmq_exchange')
        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY = local_config.get(
            'TRAINING_PROCESSOR_SERVICE', 'rabbitmq_batch_request_routing_key')
        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE = local_config.get(
            'TRAINING_PROCESSOR_SERVICE', 'rabbitmq_train_request_task_queue')
        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_URI = local_config.get(
            'TRAINING_PROCESSOR_SERVICE', 'rabbitmq_uri')
        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_USERNAME = local_config.get(
            'TRAINING_PROCESSOR_SERVICE', 'rabbitmq_username')
        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PASSWORD = local_config.get(
            'TRAINING_PROCESSOR_SERVICE', 'rabbitmq_password')
        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_SERVER = local_config.get(
            'TRAINING_PROCESSOR_SERVICE', 'rabbitmq_server')
        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PORT = local_config.getint(
            'TRAINING_PROCESSOR_SERVICE', 'rabbitmq_port')
        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_VHOST = quote_plus(
            local_config.get('TRAINING_PROCESSOR_SERVICE', 'rabbitmq_vhost'))
        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_URL = "%s%s:%s@%s:%s/%s" % (
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_URI,
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_USERNAME,
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PASSWORD,
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_SERVER,
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PORT,
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_VHOST
        )

        ###############################################################################
        # Client Options
        ###############################################################################
        self.CLASSIFICATION_SERVER = local_config.get('CLIENT', 'classification_server')
        self.CLASSIFICATION_SERVER_PORT = local_config.getint('CLIENT', 'classification_port')
        self.CLASSIFICATION_SERVER_URI = local_config.get('CLIENT', 'classification_uri')



# from tensorflow.keras.callbacks import ModelCheckpoint
import logging
import os
from datetime import datetime
from typing import Union, Any, List, Tuple, Optional

import numpy
from numpy import ndarray
from tensorflow.keras.callbacks import EarlyStopping

from .layer import DropoutLayer, DenseLayer
from .optimizers import Optimizers
from ..config.dicebox_config import DiceboxConfig
from ..connectors.filesystem_connector import FileSystemConnector
from ..connectors.sensory_service_connector import SensoryServiceConnector
from .network import Network


class DiceboxNetwork(Network):
    __accuracy: float

    # Sensory input (unique) per network - individual memory
    __fsc: FileSystemConnector  # file system connector
    __ssc: SensoryServiceConnector  # sensory service connector

    # Helper: Early stopping.
    __early_stopper = EarlyStopping(patience=25)

    ##############################################################################
    # Feature disabled until a flipper can be added and the filenames created safely.
    # Since this now runs in a container some additional considerations must be made.
    ##############################################################################
    # Checkpoint
    # filepath = "%s/weights-improvement-{epoch:02d}-{val_acc:.2f}.hdf5" % __config.WEIGHTS_DIR
    # checkpoint = ModelCheckpoint(filepath, monitor='val_acc', verbose=1, save_best_only=True, mode='max')
    # __callbacks_list = [__early_stopper, checkpoint]

    __callbacks_list = [__early_stopper]

    def __init__(self,
                 config: DiceboxConfig,
                 optimizer: Optimizers,
                 layers: List[Union[DropoutLayer, DenseLayer]] = None,
                 create_fsc: bool = True,
                 disable_data_indexing: bool = False):

        super().__init__(config=config, optimizer=optimizer, layers=layers)

        self.__accuracy: float = 0.0

        if create_fsc is True:
            logging.debug('creating a new fsc..')
            logging.info('self.config.DATA_DIRECTORY: (%s)' % self.config.DATA_DIRECTORY)
            self.__fsc = FileSystemConnector(config=config,
                                             data_directory=self.config.DATA_DIRECTORY,
                                             disable_data_indexing=disable_data_indexing)
        else:
            logging.debug('creating a new ssc..')
            self.__ssc = SensoryServiceConnector(role='client', config=self.config)

    ## Training

    def train(self, update_accuracy=False) -> float:
        if self.config.DICEBOX_COMPLIANT_DATASET is True:
            x_train, x_test, y_train, y_test = self.get_dicebox_filesystem()
        else:
            raise Exception('Unknown dataset type!  Please define, or correct.')

        self.compile()

        self.model.fit(x_train, y_train,
                       batch_size=self.config.BATCH_SIZE,
                       epochs=self.config.EPOCHS, # using early stopping, so this limit acts like a max
                       verbose=1,
                       validation_data=(x_test, y_test),
                       callbacks=[self.__early_stopper])

        score = self.model.evaluate(x_test, y_test, verbose=0)

        if update_accuracy is True:
            self.__accuracy = score

        return score[1]  # 1 is accuracy. 0 is loss.

    def get_accuracy(self) -> float:
        return self.__accuracy

    ## Prediction

    def classify(self, network_input: Any) -> ndarray:
        if self.config.DICEBOX_COMPLIANT_DATASET is True:
            x_test: ndarray = self.__get_dicebox_raw(network_input)
        else:
            logging.error("UNKNOWN DATASET (%s) passed to classify" % self.config.NETWORK_NAME)
            raise Exception("UNKNOWN DATASET (%s) passed to classify" % self.config.NETWORK_NAME)

        if self.model is None:
            logging.error('No model! Compile the network first.')
            raise Exception('No model! Compile the network first.')

        model_prediction: ndarray = self.model.predict_classes(x_test, batch_size=1, verbose=0)
        logging.info(model_prediction)

        return model_prediction

    ## Phenometype
    ## Weights Storage Functions

    def save_model_weights(self, filename: str) -> None:
        if self.model is None:
            logging.error('No model! Compile the network first.')
            raise Exception('No model! Compile the network first.')

        logging.debug('loading weights file..')
        try:
            self.model.save_weights(str(filename))  # https://github.com/keras-team/keras/issues/11269
        except Exception as e:
            logging.error('Unable to save weights file.')
            logging.error(e)
            raise e

    def load_model_weights(self, filename: str) -> None:
        if self.model is None:
            logging.error('No model! Compile the network first.')
            raise Exception('No model! Compile the network first.')

        logging.debug('loading weights file..')
        try:
            self.model.load_weights(str(filename))  # https://github.com/keras-team/keras/issues/11269
        except Exception as e:
            logging.error('Unable to load weights file.')
            logging.error(e)
            raise e

    ## Data Centric Functions

    def __get_dicebox_raw(self, raw_image_data: Any) -> ndarray:
        # TODO: variable reuse needs to be cleaned up..

        # ugh dump to file for the time being
        filename = "%s/%s" % (self.config.TMP_DIR, datetime.now().strftime('%Y-%m-%d_%H_%M_%S_%f.tmp.png'))
        with open(filename, 'wb') as f:
            f.write(raw_image_data)

        try:
            test_image_data = self.__fsc.process_image(filename)
        except:
            logging.error('Exception caught processing image data.')
            raise Exception('Exception caught processing image data.')
        finally:
            os.remove(filename)

        test_image_data = numpy.array(test_image_data)
        test_image_data = test_image_data.astype('float32')
        test_image_data /= 255

        x_test = [test_image_data]
        x_test = numpy.array(x_test)

        return x_test

    # def get_dicebox_filesystem(self) -> [ndarray, ndarray, ndarray, ndarray]:
    def get_dicebox_filesystem(self):
        noise = self.config.NOISE
        test_batch_size = self.config.TEST_BATCH_SIZE
        train_batch_size = self.config.TRAIN_BATCH_SIZE

        logging.debug('noise: %s' % noise)
        logging.debug('train_batch_size: %s' % train_batch_size)
        logging.debug('test_batch_size: %s' % test_batch_size)

        train_image_data, train_image_labels = self.__fsc.get_batch(train_batch_size, noise=noise)
        test_image_data, test_image_labels = self.__fsc.get_batch(test_batch_size, noise=noise)

        x_train = train_image_data
        x_test = test_image_data
        y_train = train_image_labels
        y_test = test_image_labels

        return x_train, x_test, y_train, y_test

    def get_dicebox_sensory_data(self) -> [ndarray, ndarray, ndarray, ndarray]:
        logging.debug('-' * 80)
        logging.debug('get_dicebox_sensory_data(self)')
        logging.debug('-' * 80)

        noise = self.config.NOISE
        train_batch_size = self.config.TRAIN_BATCH_SIZE
        test_batch_size = self.config.TEST_BATCH_SIZE

        try:
            # train_image_data, train_image_labels = Network.__fsc.get_batch(train_batch_size, noise=noise)
            train_image_data, train_image_labels = self.__ssc.get_batch(train_batch_size, noise=noise)

            logging.debug('-' * 80)
            logging.debug('train_image_data to numpy.array')
            # logging.debug(train_image_data)

            train_image_data = numpy.array(train_image_data)
            # logging.debug(train_image_data)

            logging.debug('train_image_data astype float32')
            train_image_data = train_image_data.astype('float32')
            # logging.debug(train_image_data)

            logging.debug('train_image_data /255')
            train_image_data /= 255
            # logging.debug(train_image_data)

            logging.debug('train_image_labels to numpy.array')
            train_image_labels = numpy.array(train_image_labels)
            # logging.debug(train_image_labels)
            logging.debug('-' * 80)
        except ValueError:
            logging.debug('Caught ValueError when processing training data.')
            logging.debug('failing out..')
            raise ValueError

        try:
            # test_image_data, test_image_labels = Network.__fsc.get_batch(test_batch_size, noise=noise)
            test_image_data, test_image_labels = self.__ssc.get_batch(test_batch_size, noise=noise)

            logging.debug('-' * 80)
            logging.debug('test_image_data to numpy.array')
            # logging.debug(test_image_data)

            test_image_data = numpy.array(test_image_data)
            # logging.debug(test_image_data)

            logging.debug('test_image_data astype float32')
            test_image_data = test_image_data.astype('float32')
            # logging.debug(test_image_data)

            logging.debug('test_image_data /255')
            test_image_data /= 255
            # logging.debug(test_image_data)

            logging.debug('test_image_labels to numpy.array')
            test_image_labels = numpy.array(test_image_labels)
            # logging.debug(test_image_labels)
        except ValueError:
            logging.debug('Caught ValueError when processing test data.')
            logging.debug('failing out..')
            raise ValueError

        x_train: ndarray = train_image_data
        x_test: ndarray = test_image_data
        y_train: ndarray = train_image_labels
        y_test: ndarray = test_image_labels
        return x_train, x_test, y_train, y_test



from abc import ABC
from enum import Enum
from random import choices

from typing import Tuple


class LayerType(Enum):
    DROPOUT = 'dropout'
    DENSE = 'dense'
    CONV2D = 'conv2d'  # https://www.tensorflow.org/api_docs/python/tf/keras/layers/Conv2D


def select_random_layer_type() -> LayerType:
    return choices([
        LayerType.DROPOUT,
        LayerType.DENSE,
        LayerType.CONV2D
    ])[0]


class ActivationFunction(Enum):
    ELU = 'elu'
    EXPONENTIAL = 'exponential'
    HARD_SIGMOID = 'hard_sigmoid'
    LINEAR = 'linear'
    RELU = 'relu'
    SELU = 'selu'
    SIGMOID = 'sigmoid'
    SOFTMAX = 'softmax'
    SOFTPLUS = 'softplus'
    SOFTSIGN = 'softsign'
    SWISH = 'swish'
    TANH = 'tanh'


class Conv2DPadding(Enum):
    VALID = 'valid'
    SAME = 'same'


def select_random_conv2d_padding_type() -> Conv2DPadding:
    return choices([
        Conv2DPadding.VALID,
        Conv2DPadding.SAME
    ])[0]


class Layer(ABC):
    layer_type: LayerType

    def __init__(self, layer_type: LayerType):
        self.layer_type: LayerType = layer_type


class DenseLayer(Layer):
    size: int
    activation: ActivationFunction

    def __init__(self, size: int, activation: ActivationFunction):
        super().__init__(layer_type=LayerType.DENSE)
        self.size = size
        self.activation = activation


class DropoutLayer(Layer):
    rate: float

    def __init__(self, rate: float):
        super().__init__(layer_type=LayerType.DROPOUT)
        self.rate = rate


# https://www.tensorflow.org/api_docs/python/tf/keras/layers/Conv2D
class Conv2DLayer(Layer):
    filters: int
    kernel_size: Tuple[int, int]
    strides: Tuple[int, int]
    padding: Conv2DPadding
    activation: ActivationFunction

    def __init__(self, filters: int, kernel_size: Tuple[int, int], strides: Tuple[int, int], padding: Conv2DPadding,
                 activation: ActivationFunction):
        super().__init__(layer_type=LayerType.CONV2D)
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.padding = padding
        self.activation = activation



import logging
from typing import List, Union, Any, Tuple

from tensorflow.python.keras.layers import Dropout, Dense, Conv2D, Flatten
from tensorflow.python.keras.losses import sparse_categorical_crossentropy
from tensorflow.python.keras.models import Sequential

from .layer import DenseLayer, DropoutLayer, Conv2DLayer, LayerType, ActivationFunction, Conv2DPadding
from .optimizers import Optimizers
from ..config.dicebox_config import DiceboxConfig
from ..factories.layer_factory import LayerFactory


class Network(LayerFactory):
    # genomotype (not included in the superclass via the dice)
    __optimizer: Optimizers
    __layers: List[Union[DenseLayer, DropoutLayer, Conv2DLayer]]

    # phenotype
    model: Union[Sequential, None]

    def __init__(self, config: DiceboxConfig, optimizer: Optimizers, layers: List[Union[DropoutLayer, DenseLayer, Conv2DLayer]] = None):
        super().__init__(config=config)
        self.__optimizer: Optimizers = optimizer
        if layers is not None:
            self.__layers: List[Union[DropoutLayer, DenseLayer, Conv2DLayer]] = layers
        else:
            self.__layers: List[Union[DropoutLayer, DenseLayer, Conv2DLayer]] = []
        self.model: Union[Sequential, None] = None
        self.compile()

    def add_layer(self, layer: Union[DropoutLayer, DenseLayer, Conv2DLayer]) -> None:
        self.__layers.append(layer)
        self.compile()

    def __clear_model(self):
        if self.model:
            del self.model
        self.model = None

    def compile(self) -> None:
        self.__clear_model()

        # early exist for empty layers
        if len(self.__layers) < 1:
            return

        model = Sequential()

        first_layer: bool = True
        for layer in self.__layers:
            # build and add layer
            if layer.layer_type == LayerType.DROPOUT:
                # handle dropout
                model.add(Dropout(rate=layer.rate))
            elif layer.layer_type == LayerType.DENSE:
                neurons: int = layer.size
                activation: ActivationFunction = layer.activation
                if first_layer is True:
                    first_layer = False
                    model.add(Dense(neurons, activation=activation.value, input_shape=self.config.INPUT_SHAPE))
                else:
                    model.add(Dense(neurons, activation=activation.value))
            elif layer.layer_type == LayerType.CONV2D:
                # https://www.tensorflow.org/api_docs/python/tf/keras/layers/Conv2D
                filters: int = layer.filters
                kernel_size: Tuple[int, int] = layer.kernel_size
                strides: Tuple[int, int] = layer.strides
                padding: Conv2DPadding = layer.padding
                activation: ActivationFunction = layer.activation
                logging.debug("filters=%i, kernel_size=%s, strides=%s, padding=%s, activation=%s" % (filters, kernel_size, strides, padding.value, activation.value))
                if first_layer is True:
                    first_layer = False
                    model.add(Conv2D(filters=filters, kernel_size=kernel_size, strides=strides, padding=padding.value, activation=activation.value, input_shape=self.config.INPUT_SHAPE))
                else:
                    model.add(Conv2D(filters=filters, kernel_size=kernel_size, strides=strides, padding=padding.value, activation=activation.value))
            else:
                raise

        # add final output layers..
        # https://www.machinecurve.com/index.php/2020/03/30/how-to-use-conv2d-with-keras/
        # https://jovianlin.io/cat-crossentropy-vs-sparse-cat-crossentropy/
        model.add(Flatten())
        model.add(Dense(self.config.NB_CLASSES, activation='softmax'))
        model.compile(loss=sparse_categorical_crossentropy, optimizer=self.__optimizer.value, metrics=['accuracy'])

        # return model
        self.model = model

    def get_layer(self, layer_index: int) -> Union[DenseLayer, DropoutLayer, Conv2DLayer]:
        return self.__layers[layer_index]

    def get_layers(self) -> List[Union[DenseLayer, DropoutLayer, Conv2DLayer]]:
        return self.__layers

    def get_optimizer(self) -> Optimizers:
        return self.__optimizer

    def get_input_shape(self) -> int:
        return self.config.INPUT_SHAPE

    def get_output_size(self) -> int:
        return self.config.NB_CLASSES

    def decompile(self) -> Any:
        definition = {
            'input_shape': self.config.INPUT_SHAPE,
            'output_size': self.config.NB_CLASSES,
            'optimizer': self.__optimizer.value,
            'layers': []
        }

        for i in range(0, len(self.__layers)):
            layer = self.decompile_layer(self.__layers[i])
            definition['layers'].append(layer)

        return definition



import json
import logging
import os
from typing import List

import uuid
from tqdm import tqdm

from .config.dicebox_config import DiceboxConfig
from .models.dicebox_network import DiceboxNetwork
from .optimizers.evolutionary_optimizer import EvolutionaryOptimizer
from .utils.helpers import make_sure_path_exists


class PrimordialPool:
    config: DiceboxConfig

    def __init__(self, config: DiceboxConfig) -> None:
        self.config = config

    # Updates (Trains) the set of networks passed in.
    @staticmethod
    def train_networks(networks: List[DiceboxNetwork]) -> None:
        pbar = tqdm(total=len(networks))
        for network in networks:
            accuracy: float = network.train()
            logging.info("accuracy: %f" % accuracy)
            pbar.update(1)
        pbar.close()

    @staticmethod
    def get_average_accuracy(networks: List[DiceboxNetwork]) -> float:
        """Get the average accuracy for a group of networks."""
        total_accuracy: float = 0.0
        for network in networks:
            total_accuracy += network.get_accuracy()
        return total_accuracy / len(networks)

    @staticmethod
    def print_networks(networks: List[DiceboxNetwork]) -> None:
        """Print a list of networks."""
        logging.info('-' * 80)
        for network in networks:
            logging.info(network.decompile())

    def export_population(self, population_id: str, generation: int, population: List[DiceboxNetwork]) -> None:
        output_directory: str = str(os.path.join(self.config.POPULATION_DIR, population_id, str(generation)))
        output_file: str = str(os.path.join(output_directory, 'population.json'))
        logging.info("Writing populations to: (%s)" % output_file)
        make_sure_path_exists(output_directory)

        with open(output_file, 'w') as file:
            population_genome = []
            for individual in population:
                individual_full = {
                    'accuracy': individual.get_accuracy(),
                    'genome': individual.decompile()

                }
                population_genome.append(individual_full)
            output = {
                'population': population_genome,
                'average_accuracy': self.get_average_accuracy(population)
            }
            file.write(json.dumps(output))

    # The real entry point.  Invocation of this will apply the entire process.
    def generate(self, population_file: str = None) -> str:
        population_id = str(uuid.uuid4())
        generations: int = self.config.GENERATIONS
        population_size: int = self.config.POPULATION
        logging.info("Population ID: %s" % population_id)
        logging.info('Generations: %s' % generations)
        logging.info('Population size: %s' % population_size)

        optimizer: EvolutionaryOptimizer = EvolutionaryOptimizer(config=self.config,
                                                                 retain=0.4,
                                                                 random_select=0.1,
                                                                 mutate_chance=0.2)

        # Determine if we are loading a previous population.
        if population_file is None:
            # Do not load a previous population, generate a new one.
            networks: List[DiceboxNetwork] = optimizer.create_population(population_size)
        else:
            # load the specified population.
            with open(population_file, 'r') as file:
                population_raw = json.loads(file.read())
                networks: List[DiceboxNetwork] = optimizer.create_population(size=population_size, population_definition=population_raw)

        # Evolve over the specified number of generations.
        for i in range(generations):
            logging.info("***Doing generation %d of %d***" % (i + 1, generations))
            logging.info('-' * 80)
            logging.info('Individuals in current generation')
            PrimordialPool.print_networks(networks)
            logging.info('-' * 80)

            # Train and get accuracy for networks.
            PrimordialPool.train_networks(networks)

            # Get the average accuracy for this generation.
            average_accuracy = PrimordialPool.get_average_accuracy(networks)

            # Print out the average accuracy each generation.
            logging.info("Generation average: %.2f%%" % (average_accuracy * 100))
            logging.info('-' * 80)

            logging.info('Top 5 individuals in current generation')

            # Sort our final population.
            current_networks = sorted(networks, key=lambda x: x.get_accuracy(), reverse=True)

            # Print out the top 5 networks.
            PrimordialPool.print_networks(current_networks[:5])

            self.export_population(population_id=population_id, generation=i, population=networks)

            # Evolve, except on the last iteration.
            if i != generations - 1:
                # Do the evolution.
                networks = optimizer.evolve(networks)

        # Sort our final population.
        networks = sorted(networks, key=lambda x: x.get_accuracy(), reverse=True)

        # Print out the top 5(at max) networks.
        PrimordialPool.print_networks(networks[:5])

        return population_id




import errno
import math
import os
import struct


def lucky(noise=0.0) -> bool:
    """Simple function to determine whether or not a thing occurs.

    :param noise: floating point value for noise
    :return: boolean
    """
    if noise == 1.0:
        return True
    if noise == 0.0:
        return False
    if noise > float(ord(struct.unpack('c', os.urandom(1))[0])) / 255:
        return True
    return False


def random_index(set_size: int) -> int:
    rand: float = float(ord(struct.unpack('c', os.urandom(1))[0])) / 255
    return int(math.ceil(rand * set_size))


def random_index_between(min_index: int = 0, max_index: int = 1) -> int:
    if min_index > max_index:
        raise Exception('max must be greater than or equal to the min')

    rand: float = float(ord(struct.unpack('c', os.urandom(1))[0])) / 255
    delta: int = max_index - min_index

    if ((float(ord(struct.unpack('c', os.urandom(1))[0])) / 255) >= 0.5):
        delta_offset: int = int(math.floor(rand * delta))
    else:
        delta_offset: int = int(math.ceil(rand * delta))
    return min_index + delta_offset


def dicebox_random() -> float:
    return float(ord(struct.unpack('c', os.urandom(1))[0])) / 255


# A dropout layer can is [0, 1), we can not actually use a '1' value.
def random_strict() -> float:
    new_random_number: float = dicebox_random()
    while new_random_number >= 1.0:
        new_random_number = dicebox_random()
    return new_random_number


###############################################################################
# Allows for easy directory structure creation
# https://stackoverflow.com/questions/273192/how-can-i-create-a-directory-if-it-does-not-exist
###############################################################################
def make_sure_path_exists(path) -> None:
    try:
        if os.path.exists(path) is False:
            os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise exception




import fnmatch
import logging
import os
import struct
from typing import Dict, Any, List, Union, Optional

import numpy
from PIL import Image
from numpy import ndarray, int32, float32

from ..config.dicebox_config import DiceboxConfig
from ..utils.helpers import lucky
from tensorflow.keras.preprocessing.image import img_to_array

# can be used as a foreign (sane) data set for validation of structures.
# from tensorflow import shape
# from tensorflow.keras.datasets import cifar10


class FileSystemConnector:
    """File System Connector Class"""

    config: DiceboxConfig = None

    dataset_index: dict
    category_map: Dict[Any, int]
    pixel_cache = {}

    def __init__(self, config: DiceboxConfig, data_directory: str, disable_data_indexing=False):
        self.config = config

        self.data_directory = os.path.normpath(data_directory)
        logging.info('data directory: (%s)', self.data_directory)

        if disable_data_indexing is False:

            self.dataset_index: dict = self.get_data_set()
            logging.debug('dataset_index')
            logging.debug(self.dataset_index)

            self.category_map: Dict[Any, int] = self.get_data_set_categories()
            logging.debug('CATEGORY_MAP')
            logging.debug(self.category_map)
        else:
            logging.info('File System Connector Data Indexing Disabled.')

    def get_data_set(self) -> dict:
        """
        Returns a dictionary of [k:filename, v:array of filename and category] for the entire data set.

        When we move this to python 3, there are much better libraries to handle this.  Checkout PurePath..
        """
        data_set: Dict[str, List[Union[str, Dict[str, Any]]]] = {}
        category: Dict[str, Any]
        # filename: str
        for root, _, filenames in os.walk(self.data_directory):
            for filename in fnmatch.filter(filenames, '*.png'):
                new_entry = str(os.path.join(root, filename))
                new_entry = new_entry.replace('%s%s' % (self.data_directory, os.path.sep), '')
                new_entry = os.path.normpath(new_entry)
                category, filename = new_entry.split(os.path.sep)
                # logging.info("category: (%s), filename: (%s)" % (category, filename))
                data_set[new_entry]: Union[str, Dict[str, Any]] = [filename, category]
        # logging.info(data_set)
        return data_set

    def get_data_set_categories(self) -> Dict[Any, int]:
        """Returns the dataset index as a dictionary of categories
        :return: category map as a dictionary
        """
        natural_categories = []
        category_map = {}
        value_list = self.dataset_index.values()
        # print(self.dataset_index)
        # print(value_list)
        for item in value_list:
            # logging.info(item)
            # logging.info("natural category label: (%s)" % item[1])
            natural_categories.append(item[1])
        natural_categories = sorted(set(natural_categories))
        cat_index = 0
        for cat in natural_categories:
            # logging.info("%i: %s" % (cat_index, cat))
            category_map[cat] = cat_index
            cat_index += 1
        # logging.info(category_map)
        return category_map

    def get_batch_list(self, batch_size: int) -> List[int]:
        """For a given batch size, returns a random selection of indices

        :param batch_size: integer value
        :return: array of indices in the batch size (each index appearing only once).
        """
        output: List[int] = []
        set_size: int = len(self.dataset_index)
        value_list = list(self.dataset_index.values())
        if batch_size > set_size:
            raise Exception('Max batch size: %s, but %s was specified!' % (set_size, batch_size))

        set_indices: List[int] = []
        for i in range(0, set_size):
            set_indices.append(i)

        output_list: List[int] = []
        while len(output_list) < batch_size:
            index: int = int(round((float(ord(struct.unpack('c', os.urandom(1))[0])) / 255) * (len(set_indices) - 1)))
            output_list.append(set_indices[index])
            set_indices.remove(set_indices[index])

        for i in output_list:
            output.append(value_list[i])
        return output

    def get_batch(self, batch_size: int, noise: float = 0.0) -> List[Union[ndarray, ndarray]]:
        """For a given batch size and noise level, returns a dictionary of data and labels.

        :param batch_size: integer
        :param noise: floating point number
        :return: dictionary of image data and labels
        """
        image_data: ndarray = numpy.empty((0, self.config.IMAGE_WIDTH, self.config.IMAGE_HEIGHT, 3), dtype=float32)
        image_labels: ndarray = numpy.empty((0, 1), dtype=int32)

        category_map: Dict[Any, int] = self.category_map
        batch_list: List[int] = self.get_batch_list(batch_size)

        # build file path
        for i in range(0, batch_size):
            item = batch_list[i]
            filename = "%s/%s/%s" % (self.data_directory, item[1], item[0])
            logging.debug("(%s)(%s)", category_map[item[1]], filename)
            # logging.info("  natural category label: (%s)" % item[1])
            # print("  neural network category label: (%i)" % category_map[item[1]])

            image_labels = numpy.append(image_labels, numpy.array([[category_map[item[1]]]], dtype=int32), axis=0)

            ### Cache Invalidation, hahahaha!
            # Help prevent over-fitting, and allow for new
            # sensory data to enter the cache, even when a cache
            # hit would occur.
            if filename in self.pixel_cache and not lucky(noise):
                del self.pixel_cache[filename]

            # use pixel cache if possible, [k,v] (filename, pixeldata)
            if filename in self.pixel_cache:
                # found in cache
                pixel_data = self.pixel_cache[filename]
                logging.debug("loaded cached pixel data for (%s)", filename)
            else:
                pixel_data = self.process_image(filename, noise)
                self.pixel_cache[filename] = pixel_data  # add to cache
                logging.debug("cached pixel data for (%s)", filename)

            # add new image to the set
            image_data = numpy.append(image_data, numpy.array([pixel_data], dtype=float32), axis=0)

        # Validation using another data set , note the dimensions are 32x32
        # (input_train, target_train), (input_test, target_test) = cifar10.load_data()
        # Parse numbers as floats
        # input_train = input_train.astype('float32')
        # input_test = input_test.astype('float32')
        # Scale data
        # input_train = input_train / 255
        # input_test = input_test / 255
        # return [input_train, target_train]

        return [image_data, image_labels]

    def process_image(self, filename: str, noise: float = 0.0) -> Optional[ndarray]:
        # For a given filename, and noise level, will return a numpy array of pixel data.

        local_image = Image.open(filename).convert('RGB')  # Load as RGB

        original_width, original_height = local_image.size
        # original_size = original_width, original_height
        logging.debug("original size: (%i, %i)", original_width, original_height)

        # See if there is noise
        if float(noise) > float(ord(struct.unpack('c', os.urandom(1))[0])) / 255:
            #  # then we introduce noise
            rotation_angle = float(ord(struct.unpack('c', os.urandom(1))[0])) / 255 * 10
            if float(ord(struct.unpack('c', os.urandom(1))[0])) / 255 > 0.5:
                rotation_angle = rotation_angle * -1
            logging.debug("Rotating image %f", rotation_angle)
            local_image = local_image.rotate(rotation_angle)
            logging.debug("new size: (%i, %i)", local_image.size[0], local_image.size[1])

        # Perform a random scale
        if float(noise) > float(ord(struct.unpack('c', os.urandom(1))[0])) / 255:
            # random_scale = float(ord(struct.unpack('c', os.urandom(1))[0])) / 255
            random_scale = 1.5 * float(ord(struct.unpack('c', os.urandom(1))[0])) / 255
            while random_scale < 0.7:
                random_scale = 1.5 * float(ord(struct.unpack('c', os.urandom(1))[0])) / 255

            logging.debug("Scaling image %f", random_scale)
            width, height = local_image.size
            new_width = int(width * random_scale)
            new_height = int(height * random_scale)
            if new_width > 0 and new_height > 0:
                new_size = int(width * random_scale), int(height * random_scale)
                # logging.info(new_size)
                local_image = local_image.resize(new_size)
                # local_image = local_image.resize(new_size, Image.ANTIALIAS)
                logging.debug("new size: (%i, %i)", local_image.size[0], local_image.size[1])

        # Crop Image If Required
        logging.debug('Crop image if required')

        # Now ensure we are the same dimensions as when we started
        new_width, new_height = local_image.size
        # if new_width > original_width or new_height > original_height:
        new_middle_x = float(new_width) / 2
        new_middle_y = float(new_height) / 2
        left = int(new_middle_x - float(original_width) / 2)
        upper = int(new_middle_y - float(original_height) / 2)
        right = int(new_middle_x + float(original_width) / 2)
        lower = int(new_middle_y + float(original_height) / 2)
        logging.debug("left: %i", left)
        logging.debug("upper: %i", upper)
        logging.debug("right: %i", right)
        logging.debug("lower: %i", lower)

        local_image = local_image.crop((left, upper, right, lower))
        logging.debug("new size: (%i, %i)", local_image.size[0], local_image.size[1])

        # Ensure the input will match in input tensor
        # local_image = local_image.resize((original_width, original_height), Image.ANTIALIAS)
        local_image = local_image.resize((self.config.IMAGE_WIDTH, self.config.IMAGE_HEIGHT), Image.ANTIALIAS)
        logging.debug("new size: (%i, %i)", local_image.size[0], local_image.size[1])

        # dump to file for manual review
        # filename = datetime.now().strftime('transform_%Y-%m-%d_%H_%M_%S_%f.png')
        # local_image.save("./tmp/%s" % filename)

        # https://www.tensorflow.org/api_docs/python/tf/keras/preprocessing/image/img_to_array
        image_data = img_to_array(local_image)
        return image_data



import json
import os

from .base_config import BaseConfig


class DiceboxConfig(BaseConfig):

    def __init__(self, config_file: str = 'dicebox.config'):
        super().__init__(config_file=config_file)

        ###############################################################################
        # Data Set Options
        ###############################################################################

        # Load user defined config
        if 'DATASET' in os.environ:
            self.DATASET = os.environ['DATASET']

        if 'DICEBOX_COMPLIANT_DATASET' in os.environ:
            if os.environ['DICEBOX_COMPLIANT_DATASET'] == 'True':
                self.DICEBOX_COMPLIANT_DATASET = True
            elif os.environ['DICEBOX_COMPLIANT_DATASET'] == 'False':
                self.DICEBOX_COMPLIANT_DATASET = False
            else:
                raise

        if 'NB_CLASSES' in os.environ:
            self.NB_CLASSES = int(os.environ['NB_CLASSES'])

        if 'IMAGE_WIDTH' in os.environ:
            self.IMAGE_WIDTH = int(os.environ['IMAGE_WIDTH'])

        if 'IMAGE_HEIGHT' in os.environ:
            self.IMAGE_HEIGHT = int(os.environ['IMAGE_HEIGHT'])

        if 'DATA_BASE_DIRECTORY' in os.environ:
            self.DATA_BASE_DIRECTORY = os.environ['DATA_BASE_DIRECTORY']

        ###############################################################################
        # Build Calculated Configs
        ###############################################################################
        self.NETWORK_NAME = "%s_%ix%i" % (self.DATASET, self.IMAGE_WIDTH, self.IMAGE_HEIGHT)
        self.INPUT_SHAPE = (self.IMAGE_WIDTH, self.IMAGE_HEIGHT, 3)  # 3 indicates the number of channels (RGB) so 3.
        self.DATA_DIRECTORY = "%s/%s/data/" % (self.DATA_BASE_DIRECTORY, self.NETWORK_NAME)

        ###############################################################################
        # Neural Network Taxonomy Options
        ###############################################################################
        if 'MIN_NEURONS' in os.environ:
            self.MIN_NEURONS = int(os.environ['MIN_NEURONS'])

        if 'MAX_NEURONS' in os.environ:
            self.MAX_NEURONS = int(os.environ['MAX_NEURONS'])

        if 'MIN_LAYERS' in os.environ:
            self.MIN_LAYERS = int(os.environ['MIN_LAYERS'])

        if 'MAX_LAYERS' in os.environ:
            self.MAX_LAYERS = int(os.environ['MAX_LAYERS'])

        if 'LAYER_TYPES' in os.environ:
            self.LAYER_TYPES = os.environ['LAYER_TYPES']

        if 'ACTIVATION' in os.environ:
            self.ACTIVATION = os.environ['ACTIVATION']

        if 'OPTIMIZER' in os.environ:
            self.OPTIMIZER = os.environ['OPTIMIZER']

        self.TAXONOMY = {
            'min_neurons': self.MIN_NEURONS,
            'max_neurons': self.MAX_NEURONS,
            'min_layers': self.MIN_LAYERS,
            'max_layers': self.MAX_LAYERS,
            'layer_types': json.loads(self.LAYER_TYPES),
            'activation': json.loads(self.ACTIVATION),
            'optimizer': json.loads(self.OPTIMIZER)
        }

        ###############################################################################
        # Evolution Options
        ###############################################################################
        if 'EPOCHS' in os.environ:
            self.EPOCHS = int(os.environ['EPOCHS'])

        # Number of times to evolve the population.
        if 'GENERATIONS' in os.environ:
            self.GENERATIONS = int(os.environ['GENERATIONS'])

        # Number of networks in each generation.
        if 'POPULATION' in os.environ:
            self.POPULATION = int(os.environ['POPULATION'])

        if 'NOISE' in os.environ:
            self.NOISE = float(os.environ['NOISE'])

        ###############################################################################
        # Training Options / Settings for the 1920x1080 dataset
        ###############################################################################
        if 'BATCH_SIZE' in os.environ:
            self.BATCH_SIZE = int(os.environ['BATCH_SIZE'])

        if 'TRAIN_BATCH_SIZE' in os.environ:
            self.TRAIN_BATCH_SIZE = int(os.environ['TRAIN_BATCH_SIZE'])

        if 'TEST_BATCH_SIZE' in os.environ:
            self.TEST_BATCH_SIZE = int(os.environ['TEST_BATCH_SIZE'])

        if 'LOAD_BEST_WEIGHTS_ON_START' in os.environ:
            if os.environ['LOAD_BEST_WEIGHTS_ON_START'] == 'False':
                self.LOAD_BEST_WEIGHTS_ON_START = False
            elif os.environ['LOAD_BEST_WEIGHTS_ON_START'] == 'True':
                self.LOAD_BEST_WEIGHTS_ON_START = True
            else:
                raise

        ###############################################################################
        # Direcrtory Options
        ###############################################################################
        if 'LOGS_DIR' in os.environ:
            self.LOGS_DIR = os.environ['LOGS_DIR']

        if 'WEIGHTS_DIR' in os.environ:
            self.WEIGHTS_DIR = os.environ['WEIGHTS_DIR']

        if 'TMP_DIR' in os.environ:
            self.TMP_DIR = os.environ['TMP_DIR']

        if 'POPULATION_DIR' in os.environ:
            self.POPULATION_DIR = os.environ['POPULATION_DIR']

        ###############################################################################
        # Server Options
        ###############################################################################
        if 'API_ACCESS_KEY' in os.environ:
            self.API_ACCESS_KEY = os.environ['API_ACCESS_KEY']

        if 'API_VERSION' in os.environ:
            self.API_VERSION = os.environ['API_VERSION']

        if 'LISTENING_HOST' in os.environ:
            self.LISTENING_HOST = os.environ['LISTENING_HOST']

        if 'FLASK_DEBUG' in os.environ:
            if os.environ['FLASK_DEBUG'] == 'True':
                self.FLASK_DEBUG = True
            elif os.environ['FLASK_DEBUG'] == 'False':
                self.FLASK_DEBUG = False
            else:
                raise

        if 'MODEL_WEIGHTS_FILENAME' in os.environ:
            self.MODEL_WEIGHTS_FILENAME = os.environ['MODEL_WEIGHTS_FILENAME']

        ###############################################################################
        # Sensory Service Options
        ###############################################################################
        if 'SENSORY_SERVER' in os.environ:
            self.SENSORY_SERVER = os.environ['SENSORY_SERVER']

        if 'SENSORY_PORT' in os.environ:
            self.SENSORY_PORT = int(os.environ['SENSORY_PORT'])

        if 'SENSORY_URI' in os.environ:
            self.SENSORY_URI = os.environ['SENSORY_URI']

        if 'SENSORY_SERVICE_RABBITMQ_EXCHANGE' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_EXCHANGE = os.environ['SENSORY_SERVICE_RABBITMQ_EXCHANGE']

        if 'SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_ROUTING_KEY' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_ROUTING_KEY = os.environ[
                'SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_ROUTING_KEY']

        if 'SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_TASK_QUEUE' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_TASK_QUEUE = os.environ[
                'SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_TASK_QUEUE']

        if 'SENSORY_SERVICE_RABBITMQ_URI' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_URI = os.environ['SENSORY_SERVICE_RABBITMQ_URI']

        if 'SENSORY_SERVICE_RABBITMQ_USERNAME' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_USERNAME = os.environ['SENSORY_SERVICE_RABBITMQ_USERNAME']

        if 'SENSORY_SERVICE_RABBITMQ_PASSWORD' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_PASSWORD = os.environ['SENSORY_SERVICE_RABBITMQ_PASSWORD']

        if 'SENSORY_SERVICE_RABBITMQ_SERVER' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_SERVER = os.environ['SENSORY_SERVICE_RABBITMQ_SERVER']

        if 'SENSORY_SERVICE_RABBITMQ_PORT' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_PORT = int(os.environ['SENSORY_SERVICE_RABBITMQ_PORT'])

        if 'SENSORY_SERVICE_RABBITMQ_VHOST' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_VHOST = os.environ['SENSORY_SERVICE_RABBITMQ_VHOST']

        self.SENSORY_SERVICE_RABBITMQ_URL = "%s%s:%s@%s:%s/%s" % (
            self.SENSORY_SERVICE_RABBITMQ_URI,
            self.SENSORY_SERVICE_RABBITMQ_USERNAME,
            self.SENSORY_SERVICE_RABBITMQ_PASSWORD,
            self.SENSORY_SERVICE_RABBITMQ_SERVER,
            self.SENSORY_SERVICE_RABBITMQ_PORT,
            self.SENSORY_SERVICE_RABBITMQ_VHOST
        )

        if 'SENSORY_SERVICE_SHARD_SIZE' in os.environ:
            self.SENSORY_SERVICE_SHARD_SIZE = int(os.environ['SENSORY_SERVICE_SHARD_SIZE'])

        ###############################################################################
        # Training Service Options
        ###############################################################################

        if 'TRAINING_SERVICE_RABBITMQ_EXCHANGE' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_EXCHANGE = os.environ['TRAINING_SERVICE_RABBITMQ_EXCHANGE']

        if 'TRAINING_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY = os.environ[
                'TRAINING_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY']

        if 'TRAINING_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE = os.environ[
                'TRAINING_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE']

        if 'TRAINING_SERVICE_RABBITMQ_RABBITMQ_URI' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_RABBITMQ_URI = os.environ['TRAINING_SERVICE_RABBITMQ_RABBITMQ_URI']

        if 'TRAINING_SERVICE_RABBITMQ_USERNAME' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_USERNAME = os.environ['TRAINING_SERVICE_RABBITMQ_USERNAME']

        if 'TRAINING_SERVICE_RABBITMQ_PASSWORD' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_PASSWORD = os.environ['TRAINING_SERVICE_RABBITMQ_PASSWORD']

        if 'TRAINING_SERVICE_RABBITMQ_SERVER' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_SERVER = os.environ['TRAINING_SERVICE_RABBITMQ_SERVER']

        if 'TRAINING_SERVICE_RABBITMQ_PORT' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_PORT = int(os.environ['TRAINING_SERVICE_RABBITMQ_PORT'])

        if 'TRAINING_SERVICE_RABBITMQ_VHOST' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_VHOST = os.environ['TRAINING_SERVICE_RABBITMQ_VHOST']

        self.TRAINING_SERVICE_RABBITMQ_URL = "%s%s:%s@%s:%s/%s" % (
            self.TRAINING_SERVICE_RABBITMQ_RABBITMQ_URI,
            self.TRAINING_SERVICE_RABBITMQ_USERNAME,
            self.TRAINING_SERVICE_RABBITMQ_PASSWORD,
            self.TRAINING_SERVICE_RABBITMQ_SERVER,
            self.TRAINING_SERVICE_RABBITMQ_PORT,
            self.TRAINING_SERVICE_RABBITMQ_VHOST
        )

        ###############################################################################
        # Training Processor Options
        ###############################################################################

        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_EXCHANGE' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_EXCHANGE = os.environ[
                'TRAINING_PROCESSOR_SERVICE_RABBITMQ_EXCHANGE']

        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY = os.environ[
                'TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY']

        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE = os.environ[
                'TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE']

        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_URI' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_URI = os.environ[
                'TRAINING_PROCESSOR_SERVICE_RABBITMQ_URI']

        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_USERNAME' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_USERNAME = os.environ[
                'TRAINING_PROCESSOR_SERVICE_RABBITMQ_USERNAME']

        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_PASSWORD' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PASSWORD = os.environ[
                'TRAINING_PROCESSOR_SERVICE_RABBITMQ_PASSWORD']

        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_SERVER' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_SERVER = os.environ['TRAINING_PROCESSOR_SERVICE_RABBITMQ_SERVER']

        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_PORT' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PORT = int(os.environ['TRAINING_PROCESSOR_SERVICE_RABBITMQ_PORT'])

        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_VHOST' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_VHOST = os.environ['TRAINING_PROCESSOR_SERVICE_RABBITMQ_VHOST']

        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_URL = "%s%s:%s@%s:%s/%s" % (
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_URI,
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_USERNAME,
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PASSWORD,
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_SERVER,
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PORT,
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_VHOST
        )

        ###############################################################################
        # Client Options
        ###############################################################################

        if 'CLASSIFICATION_SERVER' in os.environ:
            self.CLASSIFICATION_SERVER = os.environ['CLASSIFICATION_SERVER']

        if 'CLASSIFICATION_SERVER_PORT' in os.environ:
            self.CLASSIFICATION_SERVER_PORT = int(os.environ['CLASSIFICATION_SERVER_PORT'])

        if 'CLASSIFICATION_SERVER_URI' in os.environ:
            self.CLASSIFICATION_SERVER_URI = os.environ['CLASSIFICATION_SERVER_URI']



from typing import Any, Tuple

from .layer_factory import LayerFactory
from ..config.dicebox_config import DiceboxConfig
from ..models.layer import LayerType, ActivationFunction, Conv2DLayer, Conv2DPadding, DenseLayer, DropoutLayer
from ..models.network import Network, Optimizers
from ..utils.helpers import random_index, random_index_between


# The birthing chambers ...

class NetworkFactory(LayerFactory):
    def __init__(self, config: DiceboxConfig):
        super().__init__(config=config)

    # Processes the network which we store externally
    def create_network(self, network_definition: Any) -> Network:
        optimizer: Optimizers = Optimizers(network_definition['optimizer'])
        input_shape: int = network_definition['input_shape']
        output_size: int = network_definition['output_size']

        new_network = Network(config=self.config, optimizer=optimizer)

        if 'layers' not in network_definition:
            network_definition['layers'] = []

        # Process layers
        for layer in network_definition['layers']:
            if layer['type'] == LayerType.DENSE.value:
                size: int = layer['size']
                activation: ActivationFunction = ActivationFunction(layer['activation'])
                new_layer: DenseLayer = self.build_dense_layer(size=size, activation=activation)
                new_network.add_layer(new_layer)
            elif layer['type'] == LayerType.DROPOUT.value:
                rate: float = layer['rate']
                new_layer: DropoutLayer = self.build_dropout_layer(rate=rate)
                new_network.add_layer(new_layer)
            elif layer['type'] == LayerType.CONV2D.value:
                filters: int = layer['filters']
                kernel_size: Tuple[int, int] = layer['kernel_size']
                strides: Tuple[int, int] = layer['strides']
                padding: Conv2DPadding = Conv2DPadding(layer['padding'])
                activation: ActivationFunction = ActivationFunction(layer['activation'])
                new_layer: Conv2DLayer = Conv2DLayer(filters=filters, kernel_size=kernel_size, strides=strides, padding=padding, activation=activation)
                new_network.add_layer(new_layer)
            else:
                raise Exception("Unsupported layer type: (%s) provided." % layer['type'])

        new_network.compile()
        return new_network

    def create_random_network(self) -> Network:
        # Select an optimizer
        optimizer_index: int = random_index(len(self.config.TAXONOMY['optimizer']))
        optimizer: str = self.config.TAXONOMY['optimizer'][optimizer_index - 1]

        network: Network = Network(config=self.config,
                                   optimizer=Optimizers(optimizer))

        # Determine the number of layers..
        layer_count: int = random_index_between(self.config.TAXONOMY['min_layers'],
                                                self.config.TAXONOMY['max_layers'])
        for layer_index in range(0, layer_count):
            # add new random layer to the network
            network.add_layer(self.build_random_layer())

        network.compile()
        return network



from abc import ABC
from typing import Union, Any, Tuple

from ..config.dicebox_config import DiceboxConfig
from ..models.layer import ActivationFunction, LayerType, DropoutLayer, DenseLayer, Conv2DLayer, Conv2DPadding, select_random_conv2d_padding_type
from ..utils.helpers import random_index, random_index_between, random_strict


class LayerFactory(ABC):
    config: DiceboxConfig = None

    def __init__(self, config: DiceboxConfig):
        self.config = config

    def build_random_layer(self) -> Union[DropoutLayer, DenseLayer, Conv2DLayer]:
        # determine what the layer type will be
        layer_type_index = random_index_between(min_index = 0, max_index=len(self.config.TAXONOMY['layer_types']))
        # layer_type_index = random_index(len(self.config.TAXONOMY['layer_types']))
        layer_type = self.config.TAXONOMY['layer_types'][layer_type_index - 1]
        if layer_type == LayerType.DROPOUT.value:
            return LayerFactory.build_dropout_layer(rate=random_strict())
        elif layer_type == LayerType.DENSE.value:
            # determine the size and activation function to use.
            size: int = random_index_between(self.config.TAXONOMY['min_neurons'],
                                             self.config.TAXONOMY['max_neurons'])
            activation_index: int = random_index(len(self.config.TAXONOMY['activation']))
            activation: str = self.config.TAXONOMY['activation'][activation_index - 1]

            return LayerFactory.build_dense_layer(size=size, activation=ActivationFunction(activation))
        elif layer_type == LayerType.CONV2D.value:
            # https://www.tensorflow.org/api_docs/python/tf/keras/layers/Conv2D
            filters: int = random_index_between(1, self.config.MAX_NEURONS)  # TODO?
            kernel_size: Tuple[int, int] = (random_index_between(1, 4), random_index_between(1, 4))
            strides: Tuple[int, int] = (1, 1) # (random_index_between(1, 2), random_index_between(1, 2))

            activation_index: int = random_index(len(self.config.TAXONOMY['activation']))
            activation: str = self.config.TAXONOMY['activation'][activation_index - 1]

            return LayerFactory.build_conv2d_layer(filters=filters,
                                                   kernel_size=kernel_size,
                                                   strides=strides,
                                                   padding=select_random_conv2d_padding_type(),
                                                   activation=ActivationFunction(activation))
        else:
            raise Exception("Unsupported layer type: (%s) provided." % layer_type)

    @staticmethod
    def build_dropout_layer(rate: float) -> DropoutLayer:
        return DropoutLayer(rate=rate)

    @staticmethod
    def build_dense_layer(size: int, activation: ActivationFunction) -> DenseLayer:
        return DenseLayer(size=size, activation=activation)

    @staticmethod
    def build_conv2d_layer(filters: int, kernel_size: Tuple[int, int], strides: Tuple[int, int], padding: Conv2DPadding, activation: ActivationFunction) -> Conv2DLayer:
        return Conv2DLayer(filters=filters, kernel_size=kernel_size, strides=strides, padding=padding, activation=activation)

    @staticmethod
    def decompile_layer(layer: Union[DenseLayer, DropoutLayer, Conv2DLayer]) -> Any:
        definition = {}

        if layer.layer_type == LayerType.DROPOUT:
            definition['type'] = LayerType.DROPOUT.value
            definition['rate'] = layer.rate
        elif layer.layer_type == LayerType.DENSE:
            definition['type'] = LayerType.DENSE.value
            definition['size'] = layer.size
            definition['activation'] = layer.activation.value
        elif layer.layer_type == LayerType.CONV2D:
            definition['type'] = LayerType.CONV2D.value
            definition['filters'] = layer.filters
            definition['kernel_size'] = layer.kernel_size
            definition['strides'] = layer.strides
            definition['padding'] = layer.padding.value
            definition['activation'] = layer.activation.value
        else:
            raise

        return definition



import copy
from functools import reduce
from operator import add
from typing import List, Any, Tuple

from ..config.dicebox_config import DiceboxConfig
from ..factories.network_factory import NetworkFactory
from ..models.dicebox_network import DiceboxNetwork
from ..models.layer import select_random_conv2d_padding_type
from ..models.network import Network
from ..models.optimizers import select_random_optimizer
from ..utils.helpers import lucky, random_index, random_index_between, dicebox_random, random_strict


class EvolutionaryOptimizer(NetworkFactory):
    """Class that implements genetic algorithm for MLP optimization."""

    mutate_chance: float
    random_select: float
    retain: float

    def __init__(self,
                 config: DiceboxConfig,
                 retain: float = 0.4,
                 random_select: float = 0.1,
                 mutate_chance: float = 0.2):
        super().__init__(config=config)

        self.mutate_chance: float = mutate_chance
        self.random_select: float = random_select
        self.retain: float = retain

    def create_population(self, size: int, population_definition=None) -> List[DiceboxNetwork]:
        # Create a population of random networks.
        population: List[DiceboxNetwork] = []

        if population_definition is not None:
            # build from config...
            print('Building population from external source')
            for proto_individual in population_definition['population']:
                individual_genome = proto_individual['genome']
                individual_network: Network = self.create_network(network_definition=individual_genome)
                individual_dicebox_network = self.build_dicebox_network(network=individual_network)
                population.append(individual_dicebox_network)
        else:
            # else we build a random network.
            for _ in range(0, size):
                # Create a random network.

                random_network: Network = self.create_random_network()
                dn: DiceboxNetwork = DiceboxNetwork(config=self.config,
                                                    optimizer=random_network.get_optimizer(),
                                                    layers=random_network.get_layers())

                # Add the network to our population.
                population.append(dn)
        return population

    @staticmethod
    def fitness(network: DiceboxNetwork) -> float:
        """Return the accuracy, which is our fitness function."""
        return network.get_accuracy()

    def grade(self, population: List[DiceboxNetwork]) -> float:
        # Find average fitness for a population.
        summed: float = reduce(add, (self.fitness(network) for network in population))
        return summed / float((len(population)))

    ## This function operates on the genome of the individual, and not the higher order constructs.
    # TODO: this should include variation between the N parents as well.
    # TODO: what would it mean if the config 's came from the parents..?
    def breed(self, mother, father, offspringCount: int = 2) -> List[Any]:
        # Creates offspring
        children: List[Any] = []

        for _ in range(offspringCount):
            #
            # build our network genome
            #

            child = {
                'input_shape': self.config.INPUT_SHAPE,
                'output_size': self.config.NB_CLASSES
            }

            #
            # Pick which parent's optimization function is passed on to offspring
            #
            if lucky(0.5):
                child['optimizer'] = mother['optimizer']
            else:
                child['optimizer'] = father['optimizer']

            #
            # Determine the number of layers
            #
            if lucky(0.5):
                layer_count: int = len(mother['layers'])
            else:
                layer_count: int = len(father['layers'])

            #
            # build layers
            #
            child['layers'] = []
            for layer_index in range(0, layer_count):
                # Pick which parent's layer is passed on to the offspring
                # TODO: this should include variation between the N parents as well.
                if lucky(0.5):
                    if layer_index < len(mother['layers']):
                        layer = mother['layers'][layer_index]
                        child['layers'].append(layer)
                    elif layer_index < len(father['layers']):
                        layer = father['layers'][layer_index]
                        child['layers'].append(layer)
                    else:
                        raise Exception('impossible breeding event occurred?')
                else:
                    if layer_index < len(father['layers']):
                        layer = father['layers'][layer_index]
                        child['layers'].append(layer)
                    elif layer_index < len(mother['layers']):
                        layer = mother['layers'][layer_index]
                        child['layers'].append(layer)
                    else:
                        raise Exception('impossible breeding event occurred?')
            children.append(child)
        return children

    ## Note:
    ## This function operates on the genome of the individual, and not the higher order constructs.
    def mutate(self, individual):
        mutant = copy.deepcopy(individual)

        # this introduces chaos into the new entity
        local_noise: float = self.mutate_chance

        # TODO: possibly only of the parents types..
        # see if the optimizer is mutated
        if lucky(local_noise):
            mutant['optimizer'] = select_random_optimizer().value

        # Determine the number of layers..
        layer_count = len(mutant['layers'])

        # TODO: adjust the number of layers (its easy to remove, adding could be random)?
        # now mess around within the layers
        for index in range(0, layer_count):
            # see if the layer is mutated
            if lucky(local_noise):
                # then change the layer type
                mutant['layers'][index - 1] = self.decompile_layer(self.build_random_layer())
            else:
                layer = mutant['layers'][index - 1]

                # keep checking the individual layer attributes
                if layer['type'] == 'dropout':
                    if lucky(local_noise):
                        # mutate the dropout rate
                        layer['rate'] = random_strict()
                elif layer['type'] == 'dense':
                    if lucky(local_noise):
                        # mutate the layer size
                        layer['size'] = random_index_between(self.config.TAXONOMY['min_neurons'],
                                                             self.config.TAXONOMY['max_neurons'])
                    if lucky(local_noise):
                        # mutate activation function
                        activation_index = random_index(len(self.config.TAXONOMY['activation']))
                        layer['activation'] = self.config.TAXONOMY['activation'][activation_index - 1]
                elif layer['type'] == 'conv2d':
                    # filters
                    if lucky(local_noise):
                        layer['filters'] = random_index_between(1, self.config.MAX_NEURONS)

                    # kernel size mutate (tuple part 1)
                    if lucky(local_noise):
                        original_kernel: Tuple[int, int] = layer['kernel_size']
                        layer['kernel_size'] = (random_index_between(0, self.config.IMAGE_WIDTH), original_kernel[1])

                    # kernel size mutate (tuple part 2)
                    if lucky(local_noise):
                        original_kernel: Tuple[int, int] = layer['kernel_size']
                        layer['kernel_size'] = (original_kernel[0], random_index_between(0, self.config.IMAGE_HEIGHT))

                    # TODO - I need to review the math requirements for these combos to avoid invalid layers

                    # # strides mutate (tuple part 1)
                    # if lucky(local_noise):
                    #     original_strides: Tuple[int, int] = layer['strides']
                    #     layer['strides'] = (random_index_between(0, self.config.IMAGE_WIDTH), original_strides[1])

                    # # strides mutate (tuple part 2)
                    # if lucky(local_noise):
                    #     original_strides: Tuple[int, int] = layer['strides']
                    #     layer['strides'] = (original_strides[0], random_index_between(0, self.config.IMAGE_HEIGHT))

                    # padding mutate
                    if lucky(local_noise):
                        layer['padding'] = select_random_conv2d_padding_type().value

                    # activation mutate
                    if lucky(local_noise):
                        activation_index = random_index(len(self.config.TAXONOMY['activation']))
                        layer['activation'] = self.config.TAXONOMY['activation'][activation_index - 1]
                else:
                    raise Exception('Not yet implemented!')
        return mutant

    def evolve(self, population: List[DiceboxNetwork]) -> List[DiceboxNetwork]:
        """Evolve a population of networks."""

        # Get scores for each network.
        graded_decompiled_population: List[Tuple[float, Any]] = [(self.fitness(network), network.decompile()) for
                                                                 network in population]

        # Sort on the scores.
        ranked_population: List[Any] = [x[1] for x in
                                        sorted(graded_decompiled_population, key=lambda x: x[0], reverse=True)]

        # Get the number we want to keep for the next gen.
        retain_length: int = int(len(ranked_population) * self.retain)

        # The parents are every network we want to keep.
        parent_genomes: List[Any] = copy.deepcopy(ranked_population[:retain_length])

        # For those we aren't keeping, randomly keep some anyway.
        for individual in ranked_population[retain_length:]:
            if self.random_select > dicebox_random():
                parent_genomes.append(copy.deepcopy(individual))

        # Randomly mutate some of the networks we're keeping.
        for individual in parent_genomes:
            if lucky(self.mutate_chance):
                individual = self.mutate(individual=individual)

        # Now find out how many spots we have left to fill.
        parents_length: int = len(parent_genomes)
        desired_length: int = len(population) - parents_length
        children: List[Any] = []

        # Add children, which are bred from two remaining networks.
        while len(children) < desired_length:

            # Get a random mom and dad.
            male_index: int = random_index_between(0, parents_length - 1)
            female_index: int = random_index_between(0, parents_length - 1)

            if parents_length == 1:
                # then we do not have enough parents to breed,
                # so mutate..
                mutant: Any = self.mutate(individual=parent_genomes[0])
                children.append(mutant)
            elif parents_length < 1:
                # then there are no parents..
                # generate a random network.
                random_network: Any = self.create_random_network().decompile()
                children.append(random_network)
            else:
                # then we can bread normally
                # Assuming they aren't the same network...
                if male_index != female_index:
                    male = parent_genomes[male_index]
                    female = parent_genomes[female_index]

                    # Breed them.
                    offspring: List[Any] = self.breed(male, female)

                    # Add the children one at a time.
                    for baby in offspring:
                        # Don't grow larger than desired length.
                        if len(children) < desired_length:
                            children.append(baby)

        parent_genomes.extend(children)
        parent_networks: List[Network] = [(self.create_network(genome)) for genome in parent_genomes]
        parents: List[DiceboxNetwork] = [(self.build_dicebox_network(network=network)) for network in parent_networks]
        return parents

    def build_dicebox_network(self, network: Network) -> DiceboxNetwork:
        return DiceboxNetwork(config=self.config,
                              optimizer=network.get_optimizer(),
                              layers=network.get_layers())



from enum import Enum
from random import choices


class Optimizers(Enum):
    ADADELTA = 'adadelta'
    ADAGRAD = 'adagrad'
    ADAM = 'adam'
    ADAMAX = 'adamax'
    FTRL = 'ftrl'
    NADAM = 'nadam'
    RMSPROP = 'rmsprop'
    SGD = 'sgd'


def select_random_optimizer() -> Optimizers:
    return choices([
        Optimizers.ADADELTA,
        Optimizers.ADAGRAD,
        Optimizers.ADAM,
        Optimizers.ADAMAX,
        Optimizers.FTRL,
        Optimizers.NADAM,
        Optimizers.RMSPROP,
        Optimizers.SGD
    ])[0]



import json
import logging
from datetime import datetime

import pika
import requests
from PIL import Image

from .filesystem_connector import FileSystemConnector
from ..config.dicebox_config import DiceboxConfig
from ..utils.helpers import make_sure_path_exists


class SensoryServiceConnector:
    fsc = None  # file system connector
    interface_role = None
    config = None

    def __init__(self, role, config: DiceboxConfig):
        logging.debug('Three wise monkeys')

        self.config = config

        if self.interface_role is None:
            self.interface_role = role

        # Consuming side
        if role == 'client':
            logging.debug("[%s] client side sensory interface code goes here.", self.interface_role)

        if role == 'server':
            if self.fsc is None:
                logging.debug("[%s] creating a new __fsc..", self.interface_role)
                self.fsc = FileSystemConnector(data_directory=self.config.DATA_DIRECTORY,
                                               config=self.config)

    def get_batch(self, batch_size=0, noise=0):
        logging.debug('-' * 80)
        logging.debug("get_batch(batch_size=%i, noise=%i)" % (batch_size, noise))
        logging.debug('-' * 80)

        if self.interface_role == 'client':
            logging.debug('-' * 80)
            logging.debug('we are client')
            logging.debug('-' * 80)
            # TODO: We assemble our data through successive calls to the message service
            # or rest for small..

            outjson = {}
            outjson['batch_size'] = batch_size
            outjson['noise'] = noise

            json_data = json.dumps(outjson)
            batch_request_id = self.make_sensory_api_call('api/sensory/batch', json_data, 'POST')

            # Check the value of the response!
            if batch_request_id is None or batch_request_id == {}:
                # Then we failed to contact the sensory service, or some other error occurred..
                logging.error('Error getting a sensory batch request id!')
                raise Exception('Error getting a sensory batch request id from the sensory service!')

            batch_id = batch_request_id['batch_id']

            logging.debug(batch_id)

            # the queue will stop existing if we take too long, or we clear all the messages.
            # call the sensory service to poll and pop our messages off, it will interface with rabbitmq for us

            outjson = {}
            outjson['batch_id'] = batch_id
            json_data = json.dumps(outjson)

            image_label = []
            image_data = []

            # natural_category_list = self.get_category_map()

            response = {}
            count = 0
            # while count != batch_size:
            #     logging.debug("count: %s" % count)
            #     response = self.make_sensory_api_call('api/sensory/poll', json_data, 'POST')
            #
            #     while (len(response) < 1):
            #         time.sleep(1)
            #         response = self.make_sensory_api_call('api/sensory/poll', json_data, 'POST')
            #
            #     # logging.debug(response)
            #     # batch_item = json.load(response)
            #
            #     # lets attempt to cache to file here and convert the one-hot value to the directory structure
            #     # first convert label to one-hot value
            #     cat_index = -1
            #     one_hot_cat = response['label']
            #     for i in range(0, len(one_hot_cat)):
            #         if one_hot_cat[i] == 1:
            #             cat_index = i
            #     if cat_index < 0:
            #         logging.debug('unable to decode one hot category value')
            #         raise
            #     else:
            #         logging.debug("decoded one hot category to: (%i)" % cat_index)
            #
            #     # look up human-readable category
            #     logging.debug("cat map: (%s)" % natural_category_list)
            #     current_category = natural_category_list[str(cat_index)]
            #     logging.debug("decoded natural category: (%s)" % current_category)
            #
            #     encoded_image_data = response.json.get('data')
            #     decoded_image_data = base64.b64decode(encoded_image_data)
            #     self.sensory_store('./tmp', current_category, decoded_image_data)
            #
            #     image_label.append(response['label']) #  = numpy.append(image_label,[response['label']])
            #     image_data.append(response['data'])  # = numpy.append(image_data, [response['data']])
            #     # image_label = [response['label']]
            #     # image_data = [response['data']]
            #
            #     #logging.debug(image_label)
            #     #logging.debug(image_data)
            #
            #     count += 1

            while count < batch_size:
                logging.debug("count: %s" % count)
                new_image_data = None
                new_image_label = None

                try:
                    new_image_data, new_image_label = self.sensory_batch_poll(batch_id)

                    cat_index = -1
                    if new_image_label is not None:
                        one_hot_cat = new_image_label
                        for i in range(0, len(one_hot_cat)):
                            if one_hot_cat[i] == 1:
                                cat_index = i

                    # lets attempt to cache to file here and convert the one-hot value to the directory structure
                    # first convert label to one-hot value
                    if cat_index < 0:
                        logging.error('unable to decode one hot category value')
                        raise Exception('unable to decode one hot category value')
                    else:
                        logging.debug("decoded one hot category to: (%i)" % cat_index)

                        newimage = Image.new('L', (self.config.IMAGE_WIDTH, self.config.IMAGE_HEIGHT))  # type, size
                        newimage.putdata(new_image_data)

                        #############################################################
                        # TODO: this enables caching the images to disk, for later use
                        #############################################################
                        # image review etc. Ideally we would also attempt to PULL from cache too
                        # however this has not yet been implemented...
                        #
                        # # logging.debug('raw image decoded, dumping to file ..')
                        # ret = self.image_sensory_store(self.CONFIG.TMP_DIR, cat_index, newimage)
                        # if ret is True:
                        #     logging.debug('successfully stored to disk..')
                        # else:
                        #     logging.error('failed to store to disk!')
                        #     raise Exception('failed to store to disk!')
                        #############################################################

                        image_data.append(new_image_data)
                        image_label.append(new_image_label)
                        count += 1
                except Exception as e:
                    logging.error(e)
                    logging.debug('.')
                    # raise e

            logging.debug('-' * 80)
            logging.debug('Done receiving batch.')
            logging.debug('-' * 80)
            return image_data, image_label

            # small batch approach
            # response = self.make_sensory_api_call('api/sensory/request', json_data, 'POST')
            # image_labels = response['labels']
            # image_data = response['data']
            # return image_data, image_labels

        elif self.interface_role == 'server':
            logging.debug('-' * 80)
            logging.debug('we are server')
            logging.debug('-' * 80)
            return self.fsc.get_batch(batch_size, noise=noise)
        return None

    def make_sensory_api_call(self, end_point, json_data, call_type):
        # This should implement retry...

        headers = {
            'Content-type': 'application/json',
            'API-ACCESS-KEY': self.config.API_ACCESS_KEY,
            'API-VERSION': self.config.API_VERSION
        }

        try:
            url = "%s%s:%s/%s" % (
                self.config.SENSORY_URI, self.config.SENSORY_SERVER, self.config.SENSORY_PORT, end_point)
            # logging.debug('-' * 80)
            # logging.debug(url)
            # logging.debug('-' * 80)
            response = None
            if call_type == 'GET':
                response = requests.get(url, data=json_data, headers=headers)
            elif call_type == 'POST':
                response = requests.post(url, data=json_data, headers=headers)

            if response is not None:
                if response.status_code != 500:
                    return response.json()
        except:
            return {}
        return {}

    # This can be pulled from the File System Connector, doesn't need to be here, and isn't used anymore ..
    # TODO: temporary - we should calculate this using one of the provided methods this is really for testing the threaded-caching
    def get_category_map(self):
        jdata = {}
        with open('%s/category_map.json' % self.config.WEIGHTS_DIR) as data_file:
            raw_cat_data = json.load(data_file)
        for d in raw_cat_data:
            jdata[str(raw_cat_data[d])] = str(d)
        return jdata

    # TODO: temporary - we should calculate this using one of the provided methods this is really for testing the threaded-caching
    def image_sensory_store(self, data_dir, data_category, image_obj):
        filename = "%s" % datetime.now().strftime('%Y-%m-%d_%H_%M_%S_%f.png')
        path = "%s/%s/" % (data_dir, data_category)
        full_filename = "%s%s" % (path, filename)
        logging.debug("(%s)" % (full_filename))
        make_sure_path_exists(path)
        image_obj.save(full_filename)
        return True

    def sensory_batch_poll(self, batch_id):
        # lets try to grab more than one at a time // combine and return
        # since this is going to clients lets reduce chatter

        data = None
        label = None

        url = self.config.SENSORY_SERVICE_RABBITMQ_URL
        # logging.debug(url)
        parameters = pika.URLParameters(url)

        try:
            connection = pika.BlockingConnection(parameters=parameters)

            channel = connection.channel()

            method_frame, header_frame, body = channel.basic_get(batch_id)
            if method_frame:
                # logging.debug("%s %s %s" % (method_frame, header_frame, body))
                message = json.loads(body)
                label = message['label']
                data = message['data']
                # logging.debug(label)
                # logging.debug(data)
                channel.basic_ack(method_frame.delivery_tag)
            else:
                logging.debug('no message returned')
        except Exception as e:
            logging.warning(e)
        finally:
            connection.close()
        return data, label
